/*
  Copyright (c) 2013, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#pragma once

#include "event.hpp"
#include "queue.hpp"

#include <cstdarg>
#include <list>
#include <memory>


struct gmrf : public std::enable_shared_from_this<MMSS::node_t> {
private:
	gmrf(gmrf const&) = delete;
	gmrf& operator=(gmrf const&) = delete;

public:
	gmrf(MMSS::context_t *mmss0, const std::string &name0, unsigned rand_seed0, const std::shared_ptr<const MMSS::protocol_t> &proto0) :
		mmss(mmss0), name(name0), rand_seed(rand_seed0), proto(proto0) {}

	MMSS::context_t *mmss;

	std::string name;

	gmrf_context_t *ctx;
	std::list<std::shared_ptr<MMSS::iface_t>> interfaces;

	unsigned rand_seed;

	std::shared_ptr<const MMSS::protocol_t> proto;
};

struct gmrf_iface : public std::enable_shared_from_this<MMSS::iface_t> {
private:
	gmrf_iface(gmrf_iface const&) = delete;
	gmrf_iface& operator=(gmrf_iface const&) = delete;

public:
	gmrf_iface() = default;

	std::string name;
	gmrf_addr_t address;

	MMSS::node_t *node;
	MMSS::network_t *net;
};


namespace MMSS {

class context_t : public now_t {
public:
	timeout_queue_t<event_t> event_queue;

	context_t() : event_queue(this) {}

	void vlogf_orig(const node_t *orig, int priority, const char *format, std::va_list ap);
	void logf(int priority, const char *format, ...);
};

class config_t {
public:
	std::list<std::shared_ptr<network_t>> network;
	std::list<std::shared_ptr<node_t>> nodes;
};

class network_t {
public:
	std::string name;

	std::list<std::shared_ptr<iface_t>> interfaces;
	size_t mtu;
};


std::shared_ptr<const protocol_t> load_protocol(const char *module);

bool read_config(context_t *mmss, config_t *conf, const char *filename);

void add_iface(const std::shared_ptr<node_t> &node, const std::shared_ptr<network_t> &net, const std::string &name, const gmrf_addr_t *address);

void enqueue(context_t *mmss, const std::shared_ptr<iface_t> &source, const std::shared_ptr<iface_t> &dest, const void *data, size_t len);


namespace Config {

void add_network(context_t *mmss, config_t *conf, const char *name);

}

}
